module sd1(clk,rst,sclk,miso,mosi,cs,st,addr,secnum,datain,dataout,stb,wr,rd,ready,err,debug);

input wire clk;
input wire rst;
output reg sclk;
output reg mosi;
output reg cs;
input wire miso;
input wire [6:0] addr;
input wire [31:0] secnum;
input wire stb; // write strobe
input wire wr; // start sector write
input wire rd; // start sector read
input wire [31:0] datain;
output reg [31:0] dataout;
output reg ready;
output reg err;

output reg [7:0] st;
output reg[7:0] debug;

reg [7:0] wordcnt;
reg [31:0] writebuf[128];
reg [31:0] readbuf[128];

reg[15:0] cnt;
reg[4:0] state;
reg slowspiclk;
reg fastspiclk;
reg oldrst;
reg oldstb;
reg oldwr;
reg oldrd;
reg[56:0] cmd;
reg[48:0] res;
reg[48:0] res2;
reg[26:0] cnt2;
reg clk2;
reg [23:0] watchdog;
reg[31:0]qqqq;

initial
  begin
  cnt<=0;
  state<=0;
  mosi<=1;
  sclk<=1;
  cs<=1;
  res<=0;
  ready<=0;
  err<=0;
  watchdog<=0;
  $readmemh("wb.hex",writebuf);
  end
  
always
  begin
  st<={err,ready,1'b0,state};
  if (state==0) sclk<=1;
    else if ((state>=1) && (state<14)) sclk<=slowspiclk;
    else sclk<=fastspiclk;
  end
  
always@(posedge clk)
    qqqq<=writebuf[wordcnt];

always@(posedge clk)
  begin
  slowspiclk<=cnt[7]; //195 kHz
  fastspiclk<=cnt[0]; //25 MHz 
  dataout<=readbuf[addr];
  oldrst<=rst;
  oldstb<=stb;
  
  if ((stb==1) && (oldstb==0)) writebuf[addr]<=datain;
  
  if ((rst==1) && (oldrst==0)) //posedge reset
    begin
	 cnt<=0;
	 err<=0;
	 state<=1;
	 mosi=1;
	 cs=1;
	 end

  else // if not reset 
    begin	
  //generate err and ready
 
    if (state==14)
      begin  
  	   err<=0;
		ready<=1;
	   watchdog<=0;
	   end
    else
      begin
		ready<=0;
      if (watchdog>10_000_000) 
		  begin
		  err<=1;
		  watchdog<=10_000_001;
		  end
		else
		  watchdog<=watchdog+1;
	   end
    
  // init state machine
    if (state==0) //idle after powerup
      begin
	   cnt<=0;
	   mosi<=1;
	   cs<=1;
	   end
	 
    else if (state==1) //send 88 clocks
      begin
	   if (cnt[15:8]==88) 
	     begin
		  cmd<=56'b11111111_01_000000_00000000_00000000_00000000_00000000_10010101; //cmd0
		  cnt<=1;
		  state<=2;
		  cs=0;
        end
      else cnt<=cnt+1;
      end
  
    else if (state==2) // send cmd0
      begin
      if (cnt[7:0]==4) mosi<=cmd[55];
      if (cnt[7:0]==8)
	     begin
	     cmd<=(cmd<<1) | 56'b1;	
        end 
      if (cnt[15:8]==56)
        begin
        state<=3;
        cnt<=1;
        mosi<=1;
		  end
      else
        cnt<=cnt+1;		
	   end	
	
    else if (state==3) // wait for 0 on miso	
      begin
	   if ((cnt[7:0]==127) &&(miso==0))
	     begin
		  state<=4;
		  cnt<=128;
        res<=(res<<1) | miso;
		  end
	   else cnt<=cnt+1; 
 	   end
	 
    else if (state==4)
      begin
      if (cnt[7:0]==127) res<=(res<<1) | miso;
	   cnt<=cnt+1;
	   if (cnt[15:8]==8) // TODO! check if res=1
	     begin
  		  state<=5;
		  cnt<=0;
		  cmd<=56'b11111111_01_001000_00000000_00000000_00000001_10101010_10000111; //cmd8
 	     end 
	   end
	 
    else if (state==5) // send cmd8
      begin
      if (cnt[7:0]==4)
	     mosi<=cmd[55];
      if (cnt[7:0]==8)
	     begin
	     cmd<=(cmd<<1) | 56'b1;	
        end 
      if (cnt[15:8]==56)
        begin
        state<=6;
        cnt<=1;
        end
      else
        cnt<=cnt+1;		
	   end	
	 
    else if (state==6) // wait for 0 on miso	
      begin
	   if ((cnt[7:0]==127) &&(miso==0))
	     begin
		  state<=7;
		  cnt<=128;
		  end
	   else cnt<=cnt+1;    
      end

    else if (state==7)
      begin
      if (cnt[7:0]==127)
        res<=(res<<1) | miso;
	   cnt<=cnt+1;
	   if (cnt[15:8]==40)
	     begin
		  state<=8;
		  cnt<=0;
		  cmd<=56'hFF_77_00_00_00_00_FF;
	     end
      end

    else if (state==8) // send cmd55
      begin
      if (cnt[7:0]==4)
	     mosi<=cmd[55];
      if (cnt[7:0]==8)
	     begin
	     cmd<=(cmd<<1) | 56'b1;	
        end 
      if (cnt[15:8]==56)
        begin
        state<=9;
        cnt<=0;
        end
      else
        cnt<=cnt+1;
  	  end		 
	 
    else if (state==9) // wait for 0 on miso	
      begin
	   if ((cnt[7:0]==127) &&(miso==0))
	     begin
		  state<=10;
		  cnt<=128;
		  end
	   else cnt<=cnt+1;    
      end	 
	 
    else if (state==10)
      begin
      if (cnt[7:0]==127)
        res<=(res<<1) | miso;
	   cnt<=cnt+1;
	   if (cnt[15:8]==8) // TODO! check if res=1
	     begin
		  state<=11;
		  cnt<=1;
		  cmd<=56'hFF_69_40_00_00_00_FF; //cmd41
        end 
	   end
	 	 
    else if (state==11) // send cmd41
      begin
      if (cnt[7:0]==4)
	   mosi<=cmd[55];
      if (cnt[7:0]==8)
	     begin
	     cmd<=(cmd<<1) | 56'b1;	
        end 
      if (cnt[15:8]==56)
        begin
        state<=12;
        cnt<=1;
        end
      else
        cnt<=cnt+1;
	   end			 
 
    else if (state==12) // wait for 0 on miso	
      begin
		  if ((cnt[7:0]==127) &&(miso==0))
	     begin
		  state<=13;
		  cnt<=128;
		  end
	   else cnt<=cnt+1;    
	   end	 	 
	 
    else if (state==13)
      begin
      if (cnt[7:0]==127)
        res<=(res<<1) | miso;
	   cnt<=cnt+1;
	   if (cnt[15:8]==8) 
	     begin
		  if (res[7:0]==0)
   	    begin
		    state<=14;
		    cnt<=0;
		    cs<=1;
		    end
		  else
          begin		
		    cmd<=56'hFF_77_00_00_00_00_FF; //cmd55
		    state<=8;
		    cnt<=1;
		    end
        end 
	   end	 	 	 
	
    else if (state==14) // ready for read/write
      begin
	   if ((rd==1) && (oldrd==0)) // posedge read
        begin
		  cmd<= {16'hFF_51,secnum,8'hFF};// (56'hFF_51_00_00_00_00_FF | (secnum<<8));
		  state<=15;
		  cnt<=0;
		  cs<=0;
        end
      else if ((wr==1) && (oldwr==0)) // posedge write
	     begin
		  cmd<={16'hFF_58,secnum,8'hFF};//(56'hFF_58_00_00_00_00_FF | (secnum<<8));
		  state<=16;
		  cnt<=1;
		  cs<=0;
		  end
	   else
	     begin 
	     state<=14;
	     cnt<=0;
	     cs<=1;
	     end	
	   end	
	 
    else if (state==15) // send cmd17	 
      begin
      if (cnt[0]==0)
        begin 
        mosi<=cmd[55];
	     cmd<=(cmd<<1) | 56'b1;	
        end
      if (cnt[7:1]==56)
        begin
        state<=17;
        cnt<=1;
        end
      else
	     begin
		  state<=15;
        cnt<=cnt+1;
		  end
	   end	
	
    else if (state==17) // wait for 0 on miso	
      begin
	   if (cnt[0]==1)
        begin 
        mosi<=cmd[55];
	     res<=(cmd<<1) | miso;
        end
	   if ((cnt[0]==1) && (miso==0))
	     begin
		  state<=18;
		  cnt<=0;
		  end
	   else 
	     begin
		  cnt<=cnt+1;   
        state<=17;
        end
	   end	 	

    else if (state==18) //get response
      begin
      if (cnt[0]==1) res<=(res<<1) | miso;
	   if (cnt[7:1]==8) // TODO! check if res=1
	     begin
	     debug<=res;
		  state<=19;
		  cnt<=1;
        end 
	   else
	     begin
	  	  cnt<=cnt+1; 
		  state<=18;
		  end;
	   end

    else if (state==19) // wait for 0 on miso	=> FE sent
      begin
	   if (cnt[0]==1) res<=(res<<1) |miso;
	   if ((cnt[0]==1) && (miso==0))
	     begin
	     debug<=res;
		  state<=20;
		  cnt<=0;
	     wordcnt<=0; 
   	  end
	   else 
	     begin
	     cnt<=cnt+1;   
        state<=19;
		  end
	   end	 
	 
    else if (state==20)  
      begin
      if (cnt[0]==1) res<=(res<<1) | miso;
      if (cnt[7:1]==32) 
	     begin
	  	  cnt<=1;
        wordcnt<=wordcnt+1;
		  if (wordcnt==129)
		    begin
		    state<=14;
		    cnt<=0;
		    cs<=1;
		    end
		  else if (wordcnt<128)
  		    begin
		    readbuf[wordcnt]<=res;
		    end
        end 
      else
	     begin 
	     cnt<=cnt+1;
        state<=20;
        end		
	  end	 
	 
/// ------ write -----	 
	 
    else if (state==16) // send cmd24	 
      begin
      if (cnt[0]==0)
        begin 
        mosi<=cmd[55];
	     cmd<=(cmd<<1) | 56'b1;	
        end
      if (cnt[7:1]==56)
        begin
        state<=21;
        cnt<=0;
        end
      else
        cnt<=cnt+1;
	   end	
	
    else if (state==21) // wait for 0 on miso	
      begin
	   if (miso==0)
	     begin
		  state<=22;
		  cnt<=0;
		  end
	   else cnt<=cnt+1;   
	   end	 	

    else if (state==22) //get response
      begin
      if (cnt[0]==1)
        res<=(res<<1) | miso;
	   cnt<=cnt+1;
	   if (cnt[7:1]==8) // TODO! check if res=1
	     begin
		  state<=23;
		  cnt<=0;
		  cmd<=(48'hFF_FE_00_00_00_00 | (secnum<<8));
        end 
	   end

    else if (state==23 ) //send ---- FFFE----!!!----
      begin
      res<=32'hFF_FF_FF_FE;
	   state<=24;
	   wordcnt=0;
	   end	
 
    else if (state==24) // send writebuf
      begin
      if (cnt[0]==0)
        begin 
        mosi<=res[31];
	     res<=(res<<1) | 32'b1;	
        end
      if (cnt[7:1]==32)
	     begin
		  cnt<=0;
	     wordcnt<=wordcnt+1;
	     if (wordcnt<128) res<=qqqq; 
		  if (wordcnt==128)
		    begin
		    state<=25;
		    res<=32'hffffffff;
		    end
   	  end
	   else cnt<=cnt+1;	
      end
	 
    else if (state==25)// send ffff  
      begin
      if (cnt[0]==1) res<=(res<<1) | 32'b1;
	   if (cnt[7:1]==16) 
	     begin
		  cnt<=0;
        state<=26;
		  end
      else cnt<=cnt+1;		
	   end	

    else if (state==26) // wait for 0 on miso	
      begin
	     if ((cnt[0]==1) && (miso==0))
	     begin
		  state<=27;
		  cnt<=0;
		  end
	   else cnt<=cnt+1;   
      end
	 
    else if (state==27) //get response
      begin
      if (cnt[0]==1)
        res<=(res<<1) | miso;
	   cnt<=cnt+1;
	   if (cnt[7:1]==8) 
	     begin
		  debug<=res;
		  if (res[7:4]==4'b0101) 
	  	    begin
		    state<=28;
		    cnt<=0;
		    end
        end 
	   end
	 
    else if (state==28) //wait until ff
      begin
      if (cnt[0]==1)
        res<=(res<<1) | miso;
	   cnt<=cnt+1;
	   if (cnt[7:1]==8) 
	     begin
//		debug<=res;
	  	  if (res[7:0]==8'hFF) 
	  	    begin
		    cs<=1;
		    state<=14;
		    cnt<=0;
		    end
        end 
	   end
		
    else // error!!!
      begin
      cnt<=0;
      mosi<=1;
      cs<=1;
      state<=31;
      end	 
    end 
  end
endmodule  
	 